"""Module containing GmailHandler class. The class abstracts working with
the Gmail Api Client. Methods were taken from sample code and adopted for
use in the GmailHandler class.
"""


import httplib2
import logging

from apiclient import discovery, errors

from authentication import Authentication


class GmailHandler(object):
    """Class abstracts working with the Gmail Api Client"""
    def __init__(self, config):
        print config
        if 'CLIENT_SECRET_FILE' not in config:
            raise Exception
        client_secret_file = config['CLIENT_SECRET_FILE']

        if 'SCOPES' not in config:
            raise Exception
        scopes = config['SCOPES']

        if 'APPLICATION_NAME' not in config:
            raise Exception
        application_name = config['APPLICATION_NAME']

        self.authentication = Authentication(client_secret_file, application_name, scopes)

        self.service = self._build_service()

    def _build_service(self):
        """Build Gmail Api Client discovery service"""
        # Build credentials
        credentials = self.authentication.get_credentials()
        http = credentials.authorize(httplib2.Http())
        # Build service
        service = discovery.build('gmail', 'v1', http=http)

        return service

    def get_message(self, user_id, msg_id):
        """Get a Message with given ID.

        Args:
        service: Authorized Gmail API service instance.
        user_id: User's email address. The special value "me"
        can be used to indicate the authenticated user.
        msg_id: The ID of the Message required.

        Returns:
        A Message.
        """
        try:
            message = self.service.users().messages().get(userId=user_id, id=msg_id).execute()

        except errors.HttpError, error:
            #logger.critical("GmailApiClient error: " + str(error))
            pass
        else:
            return message

    def list_messages_matching_query(self, user_id, query=''):
        """List all Messages of the user's mailbox matching the query.

        Args:
        service: Authorized Gmail API service instance.
        user_id: User's email address. The special value "me"
        can be used to indicate the authenticated user.
        query: String used to filter messages returned.
        Eg.- 'from:user@some_domain.com' for Messages from a particular sender.

        Returns:
        List of Messages that match the criteria of the query. Note that the
        returned list contains Message IDs, you must use get with the
        appropriate ID to get the details of a Message.
        """
        try:
            response = self.service.users().messages().list(userId=user_id,
                                                   q=query).execute()
            messages = []
            if 'messages' in response:
                messages.extend(response['messages'])

            while 'nextPageToken' in response:
                page_token = response['nextPageToken']
                response = self.service.users().messages().list(userId=user_id, q=query,
                                             pageToken=page_token).execute()
                messages.extend(response['messages'])
        except errors.HttpError, error:
            #logger.critical("GmailApiClient error: " + str(error))
            pass
        else:
            return messages

    def list_messages_with_labels(self, user_id, label_ids=[]):
        """List all Messages of the user's mailbox with label_ids applied.

        Args:
        service: Authorized Gmail API service instance.
        user_id: User's email address. The special value "me"
        can be used to indicate the authenticated user.
        label_ids: Only return Messages with these labelIds applied.

        Returns:
        List of Messages that have all required Labels applied. Note that the
        returned list contains Message IDs, you must use get with the
        appropriate id to get the details of a Message.
        """
        try:
            response = self.service.users().messages().list(userId=user_id,
                                               labelIds=label_ids).execute()
            messages = []
            if 'messages' in response:
                messages.extend(response['messages'])

            while 'nextPageToken' in response:
                page_token = response['nextPageToken']
                response = self.service.users().messages().list(userId=user_id,
                                                     labelIds=label_ids,
                                                     pageToken=page_token).execute()
                messages.extend(response['messages'])

        except errors.HttpError, error:
            #logger.critical("GmailApiClient error: " + str(error))
            pass
        else:
            return messages

    def modify_message(self, user_id, msg_id, msg_labels):
        """Modify the Labels on the given Message.

        Args:
        service: Authorized Gmail API service instance.
        user_id: User's email address. The special value "me"
        can be used to indicate the authenticated user.
        msg_id: The id of the message required.
        msg_labels: The change in labels.

        Returns:
        Modified message, containing updated labelIds, id and threadId.
        """
        try:
            message = self.service.users().messages().modify(userId=user_id, id=msg_id,
                                                body=msg_labels).execute()

            label_ids = message['labelIds']

            #logger.info('MODIFIED Message ID: %s - With Label IDs %s' % (msg_id, label_ids))
        except errors.HttpError, error:
            #logger.critical("GmailApiClient error: " + str(error))
            print str(error)
        else:
            return message

    def list_labels(self, user_id):
        """Get a list all labels in the user's mailbox.

        Args:
        service: Authorized Gmail API service instance.
        user_id: User's email address. The special value "me"
        can be used to indicate the authenticated user.

        Returns:
        A list all Labels in the user's mailbox.
        """
        try:
            response = self.service.users().labels().list(userId=user_id).execute()
            labels = response['labels']
            #for label in labels:
            #    print 'Label id: %s - Label name: %s' % (label['id'], label['name'])
        except errors.HttpError, error:
            #logger.critical("GmailApiClient error: " + str(error))
            pass
        else:
            return labels

    @staticmethod
    def create_message_labels():
        """Create object to update labels.

        Returns:
        A label update object.
        """
        return {'removeLabelIds': ['Label_14'], 'addLabelIds': ['Label_13']}



# mongodao.py
"""Module contains a data access object to simplify saving data to a MongoDB
database.
"""


import logging

import pymongo

from paraturegmailfeed.exceptions import MongoDbConnectionError


logger = logging.getLogger(__name__)


class MongoDao(object):
    """Data access object for MongoDB"""
    def __init__(self, mongo_uri, db_name):
        self.client = pymongo.MongoClient(mongo_uri)
        self.db_name = db_name
        self.database = self.client[self.db_name]

    def save(self, collection, data):
        """Write single document to MongoDB database"""
        if collection == 'action':
            selector = {
                'ticketNumber': data['ticketNumber'],
                'timestamp': data['timestamp']
            }
        else:
            selector = data # Whole sale match

        try:
            upserted_id = self.database[collection].update_one(selector,
                                                           {'$set': data},
                                                           upsert=True).upserted_id
        except pymongo.errors.ConnectionFailure:
            logger.critical('MongoDB: Connection failure; check that a mongod is running')
            raise MongoDbConnectionError
        else:
            logger.info("MongoDB: Insert/update successful for: " + str(data))
            return upserted_id

    def save_many(self, collection, data):
        """Write many documents to MongoDB database"""
        upserted_ids = []
        for element in data:
            upserted_id = self.save(collection, element)
            upserted_ids.append(upserted_id)

        return upserted_ids



from __future__ import print_function
import httplib2
import logging
import os

from apiclient import discovery
import oauth2client
from oauth2client import client
from oauth2client import tools


try:
    import argparse
    flags = argparse.ArgumentParser(parents=[tools.argparser]).parse_args()
except ImportError:
    flags = None


class Authentication(object):
    def __init__(self, secret_file_path, application_name, scopes):
        self.file_path = secret_file_path
        self.application_name = application_name
        self.scopes = scopes

    def get_credentials(self):
        """Gets valid user credentials from storage.

        If nothing has been stored, or if the stored credentials are invalid,
        the OAuth2 flow is completed to obtain the new credentials.

        Returns:
            Credentials, the obtained credential.
        """
        home_dir = os.path.expanduser('~')
        credential_dir = os.path.join(home_dir, '.credentials')
        if not os.path.exists(credential_dir):
            os.makedirs(credential_dir)
        credential_path = os.path.join(credential_dir,
                                       'parature-gmail-feed.json')

        store = oauth2client.file.Storage(credential_path)
        credentials = store.get()
        if not credentials or credentials.invalid:
            flow = client.flow_from_clientsecrets(self.file_path, self.scopes)
            flow.user_agent = self.application_name
            if flags:
                credentials = tools.run_flow(flow, store, flags)
            else: # Needed only for compatibility with Python 2.6
                credentials = tools.run(flow, store)

            #print credential_path
        return credentials


if __name__ == '__main__':
    get_credentials()



"""Module containing methods to match/extract strings of pertintent data from
text."""


import re

from paraturegmailfeed.exceptions import MatchNotFound


class ParatureTextParser(object):
    """Class containing methods to parse/extract data from an email from Parature"""
    @staticmethod
    def get_ticket_number(body):
        """Returns int representing ticket number"""
        ticket_number_regex = '(?:Ticket Number:</b> )(.*?)(?:</p>)'

        match = re.search(ticket_number_regex, body)
        if match:
            ticket_number = int(match.group(1))
        else:
            raise MatchNotFound('ticket_number_regex did not return any matches!')

        return ticket_number

    @staticmethod
    def get_csr_name(body):
        """Returns str representing assigned CSR's name"""
        csr_name_regex = '(?:Assigned:</b> )(.*?)(?:</p>)'

        match = re.search(csr_name_regex, body)
        if match:
            csr_name = match.group(1)
        else:
            raise MatchNotFound('csr_name_regex did not return any matches!')

        return csr_name.strip()

    @staticmethod
    def get_action_type(subject):
        """Returns str representing the action type of the email"""
        action_type_regex = '(?:#3870-[0-9]{7}: )([a-zA-Z1-4 ]*)(?: \()'

        match = re.search(action_type_regex, subject)
        if match:
            action_type = match.group(1)
        else:
            raise MatchNotFound('action_type_regex did not return any matches!')

        return action_type.strip()

    @staticmethod
    def get_assigned_from(body):
        """Returns str representing the CSR whom assigned the ticket someone else"""
        assigned_from_regex = '(?:assigned the following ticket by )([a-zA-Z ]*)(?:\.)'

        match = re.search(assigned_from_regex, body)
        if match:
            assigned_from = match.group(1)
        else:
            assigned_from = ''
        return assigned_from



"""Module containing class to extract data from Gmail API Message response
into a standard format.
"""


import base64
from datetime import datetime


class ParatureGmailMessage(object):
    """Class modeling Parature Email sent to Gmail"""
    def __init__(self, message):
        self.subject = self._get_subject(message)
        self.body = self._get_body(message)
        self.timestamp = self._get_timestamp(message)

    @staticmethod
    def _get_subject(message):
        """Return str value of subject from Gmail API Message response"""
        headers = message['payload']['headers']

        subject = None
        for header in headers:
            if header['name'] == 'Subject':
                subject = str(header['value'])
                break

        return subject

    @staticmethod
    def _get_body(message):
        """Return str value of body from Gmail API Message response"""
        body_parts = message['payload']['parts']

        original_message = body_parts[1]
        body = original_message['body']['data']

        decoded_body = base64.urlsafe_b64decode(str(body))

        return decoded_body

    def _get_timestamp(self, message):
        """Return int value of internaldate from Gmail API Message response"""
        timestamp = message['internalDate']

        return self._convert_milliseconds_to_datetime(timestamp)

    @staticmethod
    def _convert_milliseconds_to_datetime(milliseconds):
        """Return datetime object from milliseconds"""
        time_in_seconds = int(milliseconds) / 1000.0

        return datetime.fromtimestamp(time_in_seconds)



"""Module abstracting a Parature action. Transforms a Parature Gmail message
into the 'action' the client took.
"""


from paraturegmailfeed.paratureaction.message import ParatureGmailMessage
from paraturegmailfeed.paratureaction.parser import ParatureTextParser


class ParatureAction(object):
    """
    """
    parser = ParatureTextParser

    def __init__(self, gmail_message):
        message = ParatureGmailMessage(gmail_message)

        self.timestamp = message.timestamp
        self.ticket_number = self.parser.get_ticket_number(message.body)
        self.csr_name = self.parser.get_csr_name(message.body)
        self.action_type = self.parser.get_action_type(message.subject)
        self.assigned_from = self.parser.get_assigned_from(message.body)


    def to_dict(self):
        return {
            'actionType': self.action_type,
            'assignedTo': self.csr_name,
            'assignedFrom': self.assigned_from,
            'ticketNumber': self.ticket_number,
            'timestamp': self.timestamp
        }



"""Implements configuration related object.
Source: flask.config
https://github.com/mitsuhiko/flask/blob/master/flask/config.py
"""


import os
import types


class Config(dict):
    """Create a dict object for configuration. Contains special methods to load
    from a python file.
    """
    def __init__(self, root_path):
        dict.__init__(self, {})
        self.root_path = root_path

    def from_pyfile(self, filename, silent=False):
        """Updates the values in the config from a Python file.  This function
        behaves as if the file was imported as module with the
        :meth:`from_object` function.
        :param filename: the filename of the config.  This can either be an
                         absolute filename or a filename relative to the
                         root path.
        :param silent: set to ``True`` if you want silent failure for missing
                       files.
        .. versionadded:: 0.7
           `silent` parameter.
        """
        filename = os.path.join(self.root_path, filename)
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        """Updates the values from the given object.  An object can be of one
        of the following two types:
        -   a string: in this case the object with that name will be imported
        -   an actual object reference: that object is used directly
        Objects are usually either modules or classes.
        Just the uppercase variables in that object are stored in the config.
        Example usage::
            app.config.from_object('yourapplication.default_config')
            from yourapplication import default_config
            app.config.from_object(default_config)
        You should not use this function to load the actual configuration but
        rather configuration defaults.  The actual config should be loaded
        with :meth:`from_pyfile` and ideally from a location not within the
        package because the package might be installed system wide.
        :param obj: an import name or object
        """
        if isinstance(obj, str):
            obj = import_string(obj)
        for key in dir(obj):
            if key.isupper():
                val =  getattr(obj, key)
                self[key] = val



# exceptions.py
"""Module contains custom exceptions.
"""


class Error(Exception):
    """Base-class for all exceptions raised by this package."""


class MatchNotFound(Error):
    """Regex search did not return any matches."""


class MongoDbConnectionError(Error):
    """Pymongo could not connect to mongod."""



from setuptools import setup


setup(
    name='paraturegmailfeed',
    version='0.10',
    description='ETL for Parature emails routed to Gmail account',
    url='https://github.com/joelcolucci/paraturegmailfeed',
    author='Joel Colucci',
    author_email='joelcolucci@gmail.com',
    license='MIT',
    packages=['paraturegmailfeed','paraturegmailfeed.gmailhandler', 'paraturegmailfeed.paratureaction'],
    install_requires=[
        'google-api-python-client',
        'pymongo'
    ],
    zip_safe=False
)



# feed.py
"""Module containing feed.
"""


import logging
import os

from paraturegmailfeed.config import Config
from paraturegmailfeed.gmailhandler import GmailHandler
from paraturegmailfeed.mongodao import MongoDao
from paraturegmailfeed.paratureaction import ParatureAction

# LOG_CONFIG['level'] = logging.DEBUG
# logging.basicConfig(**LOG_CONFIG)


class ParatureGmailFeed(object):
    """The ParatureGmailFeed implements a feed that fetches messages from
    the Gmail API, transforms the messages, parses them for pertintent data,
    and loads them into a MongoDB database.
    """

    # The class that is used to load configuration
    config_class = Config

    # The class that is used to connect and work with Gmail API
    gmail_handler_class = GmailHandler

    root_path = os.getcwd()
    db = None

    def __init__(self, config_filepath):
        self.config = self.config_class(self.root_path)
        self.config.from_pyfile(config_filepath)
        self.db = MongoDao(self.config['MONGO_URI'], self.config['DB_NAME'])

    def run(self):
        """Run the feed"""
        handler = self.gmail_handler_class(self.config)

        user_id = self.config['USER_ID']
        target_label = self.config['TARGET_LABEL_QUERY']

        # Fetch list of messages ids
        message_list = handler.list_messages_matching_query(user_id, target_label)
        message_ids = [msg['id'] for msg in message_list]

        for msg_id in message_ids:
            gmail_message = handler.get_message(user_id, msg_id)
            action = ParatureAction(gmail_message)

            self.db.save('action', action.to_dict())

            # Modify message (apply 'Processed' label)
            message_labels = handler.create_message_labels()
            handler.modify_message(user_id, msg_id, message_labels)
